module jp.ku6.bitmap.tga;

public import jp.ku6.bitmap.core;

import std.stream;

/++
 + TGAアクセスクラス
 +/
class TruevisionGraphicsAdapterFile : Bitmap
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  filePath = ファイルパス
	 +/
	this(string filePath)
	{
		filePath_ = filePath;
	}

	@property
	{
		uint width()
		{
			return width_;
		}

		uint height()
		{
			return height_;
		}
	}

	bool open()
	{
		// ファイルをオープン
		fileStream_ = new BufferedFile(filePath_);

		// ヘッダ情報を読み出し
		ubyte[18] buffer;
		fileStream_.read(buffer);

		// イメージID非対応
		if (buffer[0] != 0)
			goto OPEN_FAILED;

		// カラーマップ非対応
		if (buffer[1] != 0)
			goto OPEN_FAILED;

		// 無圧縮RGBと圧縮RGBのみ対応
		switch (buffer[2])
		{
		case ColorType.RGB:
			colorType_ = ColorType.RGB;
			break;
		case ColorType.COMPRESSED_RGB:
			colorType_ = ColorType.COMPRESSED_RGB;
			break;
		default:
			goto OPEN_FAILED;
		}

		// カラーマップ非対応 0以外の原点非対応
		for (uint i = 3; i < 12; ++i)
			if (buffer[i] != 0)
				goto OPEN_FAILED;

		// 幅を取得
		width_ = buffer[12] | (buffer[13] << 8);

		// 高さを取得
		height_ = buffer[14] | (buffer[15] << 8);

		// ビット深度を取得
		if (buffer[16] != 24 && buffer[16] != 32)
			goto OPEN_FAILED;
		bitPerPixel_ = buffer[16];

		// 形式を取得
		flags_ = buffer[17];

		return true;

// 失敗時にはクローズしてから終わる
OPEN_FAILED:

		fileStream_.close();

		return false;
	}

	void close()
	{
		fileStream_.close();
	}

	uint[] readAsTrueColor()
	{
		uint[] destination = new uint[width_ * height_];
		return readAsTrueColor(destination, width_, height_);
	}

	uint[] readAsTrueColor(uint[] destination, uint destinationWidth, uint destinationHeight)
	{
		// 展開先バッファに展開できるサイズがあるか確認
		if (destination is null || destination.length == 0 || destinationWidth == 0 || destinationHeight == 0 || destination.length < destinationWidth * destinationHeight || destinationWidth < width_ || destinationHeight < height_)
			return null;

		// ファイルから変換元データ等をロード
		ubyte[] source = loadAndDecode();
		if (source is null)
			return null;

		if (bitPerPixel_ == 24)
		{
			ubyte* read = source.ptr;
			for (uint y; y < height_; ++y)
			{
				for (uint x; x < width_; ++x)
				{
					uint color = (read[0] << 16) | (read[1] << 8) | read[2] | 0xff000000;
					destination[y * destinationWidth + x] = color;
					read += 3;
				}
			}
		}
		else if (bitPerPixel_ == 32)
		{
			ubyte* read = source.ptr;
			for (uint y; y < height_; ++y)
			{
				for (uint x; x < width_; ++x)
				{
					uint color = (read[0] << 16) | (read[1] << 8) | read[2] | (read[3] << 24);
					destination[y * destinationWidth + x] = color;
					read += 4;
				}
			}
		}

		return destination;
	}

private:

	// 色形式
	enum ColorType : ubyte
	{
		RGB = 2,
		COMPRESSED_RGB = 10,
	}

	// ファイルパス
	string filePath_;

	// ファイルストリーム
	BufferedFile fileStream_;

	// 1ピクセルあたりのビット数
	ubyte bitPerPixel_;

	// 画像の幅
	uint width_;

	// 画像の高さ
	uint height_;

	// 画像の色形式
	ColorType colorType_;

	// フラグ
	ubyte flags_;

	static
	void horizontalReverse(ubyte* buffer, uint width, uint height, uint bytePerPixel)
	{
		uint loop = width / 2;

		for (uint y; y < height; ++y)
		{
			ubyte* left = &buffer[y * width * bytePerPixel];
			for (uint x; x < loop; ++x)
			{
				for (uint i; i < bytePerPixel; ++i)
				{
					ubyte tmp = left[x * bytePerPixel + i];
					left[x * bytePerPixel + i] = left[(width - x - 1) * bytePerPixel + i];
					left[(width - x - 1) * bytePerPixel + i] = tmp;
				}
			}
		}
	}

	static
	void verticalReverse(ubyte* buffer, uint width, uint height, uint bytePerPixel)
	{
		uint pitch = width * bytePerPixel;
		height /= 2;

		for (uint y; y < height; ++y)
		{
			ubyte* top = &buffer[y * pitch];
			ubyte* bottom = &buffer[(height * 2 - y - 1) * pitch];

			for (uint x; x < pitch; ++x)
			{
				ubyte tmp = top[x];
				top[x] = bottom[x];
				bottom[x] = tmp;
			}
		}
	}

	/++
	 + ファイルからデータをロード
	 +
	 + ファイルから必要なデータをロードする
	 + また、圧縮データの場合解凍も行う
	 +
	 + Returns:
	 +  読み込まれたデータ
	 +  失敗時にはnullを返す
	 +/
	ubyte[] loadAndDecode()
	{
		// ヘッダの直後から読み込みを開始
		fileStream_.seek(18, SeekPos.Set);

		// 1ピクセルのバイト数を取得
		uint bytePerPixel = bitPerPixel_ / 8;

		// データを読み込む領域を取得
		ubyte[] buffer = new ubyte[width_ * height_ * bytePerPixel];

		// 無圧縮カラー
		if (colorType_ == ColorType.RGB)
		{
			fileStream_.read(buffer);
		}
		// 圧縮カラー
		else if (colorType_ == ColorType.COMPRESSED_RGB)
		{
			uint counter;
			while (counter < buffer.length)
			{
				// フラグを取得
				ubyte flag;
				fileStream_.read(flag);

				if ((flag & 0x80) == 0)
				{
					// 繰り返し無しピクセル数を取得
					flag &= 0x7f;
					flag += 1;

					// 繰り返し無しデータを読み込み
					fileStream_.read(buffer[counter .. counter + flag * bytePerPixel]);

					// 読み込み済みサイズを更新
					counter += flag * bytePerPixel;
				}
				else
				{
					// 繰り返しピクセル数を取得
					flag &= 0x7f;
					flag += 1;

					// 1ピクセルのデータを読み込み
					fileStream_.read(buffer[counter .. counter + bytePerPixel]);

					// 繰り返し
					for (uint i; i < flag; ++i)
					{
						for (uint j; j < bytePerPixel; ++j)
							buffer[counter + i * bytePerPixel + j] = buffer[j + counter];
					}

					// 読み込み済みサイズを更新
					counter += flag * bytePerPixel;
				}
			}
		}

		if ((flags_ & 0x10) != 0)
			horizontalReverse(buffer.ptr, width_, height_, bytePerPixel);

		if ((flags_ & 0x20) != 1)
			verticalReverse(buffer.ptr, width_, height_, bytePerPixel);

		return buffer;
	}
}
