module jp.ku6.pmdloader.pmdloader;

import std.math;
import jp.ku6.windows.win32api.windows;
import jp.ku6.windows.directx.dxgiformat;
import jp.ku6.windows.directx.dxgitype;
import jp.ku6.windows.directx.dxgi;
import jp.ku6.windows.directx.d3dcommon;
import jp.ku6.windows.directx.d3d11;
import jp.ku6.windows.directx.d3dcompiler;
import jp.ku6.windows.game.application;
import jp.ku6.windows.game.input;
import jp.ku6.pmdloader.DepthRenderTechnique;
import jp.ku6.pmdloader.math;
import jp.ku6.pmdloader.pmd;

/++
 + 定数バッファ
 +/
struct Constants
{
	Matrix worldMatrix;
	Matrix worldCameraViewProjectionMatrix;
	Matrix worldLightViewProjectionMatrix;
}

class PmdLoader
{
	/++
	 + コンストラクタ
	 +
	 + アプリケーション起動時に一度だけ呼び出される
	 +/
	this(string filePath)
	{
		//--------------------------------------------------------------
		// 使用するDLLの読み込み

		// Direct3D 11
		d3d11Dll_ = LoadLibrary("d3d11.dll\0".ptr);
		if (d3d11Dll_ is null)
			throw new Error("Failed load d3d11.dll");

		// shader compiler
		d3dcompiler_ = LoadLibrary(D3DCOMPILER_DLL.ptr);
		if (d3dcompiler_ is null)
		{
			cleanup();
			throw new Error("Failed load d3dcompiler.dll");
		}

		//--------------------------------------------------------------
		// 基本となるDirect 3D Deviceとスワップチェインを用意

		// D3D11CreateDeviceAndSwapChainのアドレスを取得
		PFN_D3D11_CREATE_DEVICE_AND_SWAP_CHAIN D3D11CreateDeviceAndSwapChain = cast(PFN_D3D11_CREATE_DEVICE_AND_SWAP_CHAIN)GetProcAddress(d3d11Dll_, "D3D11CreateDeviceAndSwapChain\0".ptr);
		if (D3D11CreateDeviceAndSwapChain is null)
		{
			cleanup();
			throw new Error("Failed get D3D11CreateDeviceAndSwapChain function");
		}

		// Swap chainの設定
		DXGI_SWAP_CHAIN_DESC swapChainDesc;
		swapChainDesc.BufferCount = 1;
		swapChainDesc.BufferDesc.Width = Application.instance.window.clientWidth;
		swapChainDesc.BufferDesc.Height = Application.instance.window.clientHeight;
		swapChainDesc.BufferDesc.Format = DXGI_FORMAT.DXGI_FORMAT_R8G8B8A8_UNORM;
		swapChainDesc.BufferDesc.RefreshRate.Numerator = 60;
		swapChainDesc.BufferDesc.RefreshRate.Denominator = 1;
		swapChainDesc.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;
		swapChainDesc.OutputWindow = Application.instance.window.handle;
		swapChainDesc.SampleDesc.Count = 1;
		swapChainDesc.SampleDesc.Quality = 0;
		swapChainDesc.Windowed = TRUE;

		// IDXGISwapChain, ID3D11Device, ID3D11DeviceContextを作成
		D3D_FEATURE_LEVEL featureLevel;
		if (FAILED(D3D11CreateDeviceAndSwapChain(null, D3D_DRIVER_TYPE.D3D_DRIVER_TYPE_HARDWARE, null, 0, null, 0, D3D11_SDK_VERSION, &swapChainDesc, &swapChain_, &device_, &featureLevel, &immediateContext_)))
		{
			cleanup();
			throw new Error("Failed create device");
		}

		//--------------------------------------------------------------
		// 深度バッファを用意

		// 深度バッファ作成
		D3D11_TEXTURE2D_DESC textureDesc;
		textureDesc.Width  = swapChainDesc.BufferDesc.Width;
		textureDesc.Height = swapChainDesc.BufferDesc.Height;
		textureDesc.MipLevels = 1;
		textureDesc.ArraySize = 1;
		textureDesc.Format = DXGI_FORMAT.DXGI_FORMAT_D32_FLOAT;
		textureDesc.SampleDesc = swapChainDesc.SampleDesc;
		textureDesc.Usage = D3D11_USAGE.D3D11_USAGE_DEFAULT;
		textureDesc.BindFlags = D3D11_BIND_FLAG.D3D11_BIND_DEPTH_STENCIL;
		textureDesc.CPUAccessFlags = 0;
		textureDesc.MiscFlags = 0;

		ID3D11Texture2D depthBuffer;
		if (FAILED(device_.CreateTexture2D(&textureDesc, null, &depthBuffer)))
		{
			cleanup();
			throw new Error("Failed create depth buffer.");
		}

		// 深度ステンシルビューの作成
		D3D11_DEPTH_STENCIL_VIEW_DESC depthStencilViewDesc;
		depthStencilViewDesc.Format = textureDesc.Format;
		depthStencilViewDesc.ViewDimension = D3D11_DSV_DIMENSION.D3D11_DSV_DIMENSION_TEXTURE2DMS;
		depthStencilViewDesc.Flags = 0;

		if (FAILED(device_.CreateDepthStencilView(depthBuffer, &depthStencilViewDesc, &depthStencilView_)))
		{
			depthBuffer.Release();
			cleanup();
			throw new Error("Failed create depth stencil view.");
		}

		depthBuffer.Release();


		//--------------------------------------------------------------
		// シェーダコンパイルの準備

		// D3DCompileのアドレスを取得
		D3DCompileFunction D3DCompile = cast(D3DCompileFunction)GetProcAddress(d3dcompiler_, "D3DCompile\0".ptr);
		if (D3DCompile is null)
		{
			cleanup();
			throw new Error("Failed get D3DCompile function");
		}

		// シェーダコンパイル後のデータとエラーを格納する
		ID3DBlob compiled;
		ID3DBlob errors;

		//--------------------------------------------------------------
		// Input Assembler ステージで使用するデータを準備

		// Vertex shader プログラムソースをコンパイル
		string vertexShaderSource = `
		cbuffer Constants : register(b0)
		{
			float4x4 worldMatrix;
			float4x4 worldCameraViewProjectionMatrix;
			float4x4 worldLightViewProjectionMatrix;
		};

		struct VS_INPUT
		{
			float4 pos : POSITION;
			float3 nor : NORMAL;
			float2 uv : TEXCOORD;
		};

		struct PS_INPUT
		{
			float4 pos : SV_POSITION;
			float4 lpos : POSITION;
			float3 nor : NORMAL;
			float2 uv : TEXCOORD;
		};

		PS_INPUT VS_main(VS_INPUT input)
		{
			PS_INPUT output;
			output.nor = mul(input.nor, worldMatrix);
			output.pos = mul(input.pos, worldCameraViewProjectionMatrix);
			output.lpos = mul(input.pos, worldLightViewProjectionMatrix);
			output.uv = input.uv;
			return output;
		}
		`;
		if (FAILED(D3DCompile(vertexShaderSource.ptr, vertexShaderSource.length, null, null, null, "VS_main\0".ptr, "vs_4_0\0".ptr, 0, 0, &compiled, &errors)))
		{
			string message = "";
			if (errors !is null)
			{
				message = (cast(char*)errors.GetBufferPointer())[0 .. errors.GetBufferSize()].idup;
				errors.Release();
			}
			throw new Error("Failed compile vertex shader: " ~ message);
		}

		// Input assembler ステージの入力レイアウトを作成
		D3D11_INPUT_ELEMENT_DESC[3] vertexDescs;
		vertexDescs[0] = D3D11_INPUT_ELEMENT_DESC("POSITION\0".ptr, 0, DXGI_FORMAT.DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D11_INPUT_CLASSIFICATION.D3D11_INPUT_PER_VERTEX_DATA, 0);
		vertexDescs[1] = D3D11_INPUT_ELEMENT_DESC("NORMAL\0".ptr, 0, DXGI_FORMAT.DXGI_FORMAT_R32G32B32_FLOAT, 0, 12, D3D11_INPUT_CLASSIFICATION.D3D11_INPUT_PER_VERTEX_DATA, 0);
		vertexDescs[2] = D3D11_INPUT_ELEMENT_DESC("TEXCOORD\0".ptr, 0, DXGI_FORMAT.DXGI_FORMAT_R32G32_FLOAT, 0, 24, D3D11_INPUT_CLASSIFICATION.D3D11_INPUT_PER_VERTEX_DATA, 0);

		if (FAILED(device_.CreateInputLayout(vertexDescs.ptr, vertexDescs.length, compiled.GetBufferPointer(), compiled.GetBufferSize(), &inputLayout_)))
		{
			compiled.Release();
			if (errors !is null)
				errors.Release();

			cleanup();
			throw new Error("Failed create input layout");
		}


		//--------------------------------------------------------------
		// Vertex Shader ステージで使用するデータを準備

		// Vertex shaderを作成
		if (FAILED(device_.CreateVertexShader(compiled.GetBufferPointer(), compiled.GetBufferSize(), null, &vertexShader_)))
		{
			compiled.Release();
			if (errors !is null)
				errors.Release();
			cleanup();
			throw new Error("Failed create vertex shader");
		}

		compiled.Release();
		if (errors !is null)
			errors.Release();


		// 頂点シェーダ定数バッファを作成
		D3D11_BUFFER_DESC vertexShaderConstantsBufferDesc = D3D11_BUFFER_DESC(Constants.sizeof, D3D11_USAGE.D3D11_USAGE_DEFAULT, D3D11_BIND_FLAG.D3D11_BIND_CONSTANT_BUFFER, 0, 0);
		if (FAILED(device_.CreateBuffer(&vertexShaderConstantsBufferDesc, null, &vertexShaderConstantsBuffer_)))
		{
			cleanup();
			throw new Error("Failed create constants buffer");
		}

		//--------------------------------------------------------------
		// Rasterizer ステージで使用するデータを準備

		// Viewportの設定
		viewPort_.Width = swapChainDesc.BufferDesc.Width;
		viewPort_.Height = swapChainDesc.BufferDesc.Height;
		viewPort_.MinDepth = 0.0f;
		viewPort_.MaxDepth = 1.0f;
		viewPort_.TopLeftX = 0;
		viewPort_.TopLeftY = 0;

		// Rasterizer state
		D3D11_RASTERIZER_DESC rasterizerDesc;
		rasterizerDesc.FillMode = D3D11_FILL_MODE.D3D11_FILL_SOLID;
		rasterizerDesc.CullMode = D3D11_CULL_MODE.D3D11_CULL_NONE;
		rasterizerDesc.DepthClipEnable = FALSE;
		rasterizerDesc.MultisampleEnable = FALSE;
		rasterizerDesc.DepthBiasClamp = 0;
		rasterizerDesc.SlopeScaledDepthBias = 0;
		if (FAILED(device_.CreateRasterizerState(&rasterizerDesc, &rasterizerState_)))
		{
			cleanup();
			throw new Error("Failed create rasterizer state");
		}


		//--------------------------------------------------------------
		// Pixel Shader ステージで使用するデータを準備

		// テクスチャを使用しないPixel shaderを準備
		string polygonPixelShaderSource =
		`
			cbuffer Material : register(b0)
			{
				float3 diffuseColor;
				float alpha;
				float3 specularColor;
				float shininess;
				float3 ambientColor;
				float dummy;
			};

			Texture2D depthTex : register(t1);

			SamplerState sam : register(s0);

			struct PS_INPUT
			{
				float4 pos : SV_POSITION;
				float4 lpos : POSITION;
				float3 normal : NORMAL;
			    float2 uv : TEXCOORD;
			};

			float4 PS_main(PS_INPUT input) : SV_Target
			{
				// 視線ベクトル (ひとまず固定値で)
				static const float3 cameraDirection = {0.0f, 0.0f, 1.0f};

				// 平行光源向き
				static const float3 directionalLightDirection = {0.0f, -1.0f, 1.0f};

				// 平行光源色
				static const float3 directionalLightColor = {0.7f, 0.7f, 0.7f};

				// 環境光色
				static const float3 ambientLightColor = {0.8f, 0.8f, 0.8f};

				if (alpha == 0)
					discard;

				// 法線ベクトルと視線ベクトルの角度が90度より小さければ法線ベクトルを反転させる(裏面描画のため)
				float angle = dot(normalize(input.normal), normalize(cameraDirection));
				if (angle > 0)
					input.normal = -input.normal;

				// X軸、Y軸共に-1 ～ 1なのでテクスチャ座標に変換する
				float2 depthTexCoord;
				depthTexCoord.x = (input.lpos.x + 1) * 0.5;
				depthTexCoord.y = (1 - input.lpos.y) * 0.5;

				// テクスチャからZ値を拾う(最前面のZ値) 誤差調整のためいくらか後ろにずらす
				float minDepth = depthTex.Sample(sam, depthTexCoord).r + 0.0005f;

				float fraction;

				// ディレクショナルライトをを遮ってなければ計算する
				float3 diffuse = float3(0.0f, 0.0f, 0.0f);
				if (minDepth >= input.lpos.z)
				{
					// 角度がつくことによって光線密度が減った分を計算する
					fraction = dot(normalize(input.normal), normalize(directionalLightDirection));

					// 光線ベクトルと面法線ベクトルの角度が90～180度の場合のみ有効。180の時MAX
					fraction = max(-fraction, 0.0f);

					// ライトの色にかける
					diffuse = mul(directionalLightColor, fraction);

					// 素材の反射率とかける
					diffuse = diffuse * diffuseColor;
				}

				// スペキュラも影の影響をうける
				float3 specular = float3(0.0f, 0.0f, 0.0f);
				if (shininess > 0.0f && minDepth >= input.lpos.z)
				{
					// 光線の面に対する反射ベクトルを取得
					float3 reflection = reflect(normalize(directionalLightDirection), normalize(input.normal));

					// カメラと光線反射ベクトルの角度を取得
					fraction = dot(normalize(cameraDirection), normalize(reflection));

					// カメラと光線反射ベクトルの角度が90～180度の場合のみ有効。180の時MAX
					if (fraction < 0.0f)
					{
						// 光沢度の数値乗する
						fraction = pow(-fraction, shininess);

						// 計算結果の割合をライトの色にかける
						specular = mul(directionalLightColor, fraction);

						// 素材の反射率とかける
						specular = specular * specularColor;
					}
				}


				// 環境光に素材の環境光反射をかける
				float3 ambient = ambientColor * ambientLightColor;

				// 全部の光を合わせてテクスチャの色とかける
				return float4(ambient + diffuse + specular, alpha);
			}
		`;
		if (FAILED(D3DCompile(polygonPixelShaderSource.ptr, polygonPixelShaderSource.length, null, null, null, "PS_main\0".ptr, "ps_4_0\0".ptr, 0, 0, &compiled, &errors)))
		{
			string message = "";
			if (errors !is null)
			{
				message = (cast(char*)errors.GetBufferPointer())[0 .. errors.GetBufferSize()].idup;
				errors.Release();
			}
			cleanup();
			throw new Error("Failed compile pixel shader: " ~ message);
		}

		if (FAILED(device_.CreatePixelShader(compiled.GetBufferPointer(), compiled.GetBufferSize(), null, &polygonPixelShader_)))
		{
			compiled.Release();
			if (errors !is null)
				errors.Release();
			cleanup();
			throw new Error("Failed create pixel shader");
		}

		compiled.Release();
		if (errors !is null)
			errors.Release();


		// テクスチャを使用するPixel shaderを準備
		string texturePixelShaderSource =
		`
			cbuffer Material : register(b0)
			{
				float3 diffuseColor;
				float alpha;
				float3 specularColor;
				float shininess;
				float3 ambientColor;
				float dummy;
			};

			Texture2D tex : register(t0);
			Texture2D depthTex : register(t1);

			SamplerState sam : register(s0);

			struct PS_INPUT
			{
				float4 pos : SV_POSITION;
				float4 lpos : POSITION;
				float3 normal : NORMAL;
			    float2 uv : TEXCOORD;
			};

			float4 PS_main(PS_INPUT input) : SV_Target
			{
				// 視線ベクトル (ひとまず固定値で)
				static const float3 cameraDirection = {0.0f, 0.0f, 1.0f};

				// 平行光源向き
				static const float3 directionalLightDirection = {0.0f, -1.0f, 1.0f};

				// 平行光源色
				static const float3 directionalLightColor = {0.7f, 0.7f, 0.7f};

				// 環境光色
				static const float3 ambientLightColor = {0.8f, 0.8f, 0.8f};

				// モデルのアルファ値が0ならば表示しない
				if (alpha == 0)
					discard;

				// テクスチャから取ってアルファ値が0ならば表示しない
				float4 pix = tex.Sample(sam, input.uv);
				if (pix.a == 0)
					discard;

				// 法線ベクトルと視線ベクトルの角度が90度以内ならば法線ベクトルを反転させる(裏面描画のため)
				float angle = dot(normalize(input.normal), normalize(cameraDirection));
				if (angle > 0)
					input.normal = -input.normal;

				// X軸、Y軸共に-1 ～ 1なのでテクスチャ座標に変換する
				float2 depthTexCoord;
				depthTexCoord.x = (input.lpos.x + 1) * 0.5;
				depthTexCoord.y = (1 - input.lpos.y) * 0.5;

				// テクスチャからZ値を拾う(最前面のZ値) 誤差調整のためいくらか後ろにずらす
				float minDepth = depthTex.Sample(sam, depthTexCoord).r + 0.0005f;


				float fraction;

				// ディレクショナルライトをを遮ってなければ計算する
				float3 diffuse = float3(0.0f, 0.0f, 0.0f);
				if (minDepth >= input.lpos.z)
				{
					// 角度がつくことによって光線密度が減った分を計算する
					fraction = dot(normalize(input.normal), normalize(directionalLightDirection));

					// 光線ベクトルと面法線ベクトルの角度が90～180度の場合のみ有効。180の時MAX
					fraction = max(-fraction, 0.0f);

					// ライトの色にかける
					diffuse = mul(directionalLightColor, fraction);

					// 素材の反射率とかける
					diffuse = diffuse * diffuseColor;
				}


				// スペキュラも影の影響をうける
				float3 specular = float3(0.0f, 0.0f, 0.0f);
				if (shininess > 0.0f && minDepth >= input.lpos.z)
				{
					// 光線の面に対する反射ベクトルを取得
					float3 reflection = reflect(normalize(directionalLightDirection), normalize(input.normal));

					// カメラと光線反射ベクトルの角度を取得
					fraction = dot(normalize(cameraDirection), normalize(reflection));

					// カメラと光線反射ベクトルの角度が90～180度の場合のみ有効。180の時MAX
					if (fraction < 0.0f)
					{
						// 光沢度の数値乗する
						fraction = pow(-fraction, shininess);

						// 計算結果の割合をライトの色にかける
						specular = mul(directionalLightColor, fraction);

						// 素材の反射率とかける
						specular = specular * specularColor;
					}
				}


				// 環境光に素材の環境光反射をかける
				float3 ambient = ambientColor * ambientLightColor;

				// 全部の光を合わせてテクスチャの色とかける
				return float4(ambient + diffuse + specular, alpha) * pix;
			}
		`;
		if (FAILED(D3DCompile(texturePixelShaderSource.ptr, texturePixelShaderSource.length, null, null, null, "PS_main\0".ptr, "ps_4_0\0".ptr, 0, 0, &compiled, &errors)))
		{
			string message = "";
			if (errors !is null)
			{
				message = (cast(char*)errors.GetBufferPointer())[0 .. errors.GetBufferSize()].idup;
				errors.Release();
			}
			cleanup();
			throw new Error("Failed compile pixel shader: " ~ message);
		}

		if (FAILED(device_.CreatePixelShader(compiled.GetBufferPointer(), compiled.GetBufferSize(), null, &texturePixelShader_)))
		{
			compiled.Release();
			if (errors !is null)
				errors.Release();
			cleanup();
			throw new Error("Failed create pixel shader");
		}

		compiled.Release();
		if (errors !is null)
			errors.Release();

		// ピクセルシェーダ定数バッファを作成
		D3D11_BUFFER_DESC pixelShaderConstantsBufferDesc = D3D11_BUFFER_DESC(PmdFile.Material.sizeof, D3D11_USAGE.D3D11_USAGE_DEFAULT, D3D11_BIND_FLAG.D3D11_BIND_CONSTANT_BUFFER, 0, 0);
		if (FAILED(device_.CreateBuffer(&pixelShaderConstantsBufferDesc, null, &pixelShaderConstantsBuffer_)))
		{
			cleanup();
			throw new Error("Failed create constants buffer");
		}

		D3D11_SAMPLER_DESC samplerDesc;
		samplerDesc.Filter = D3D11_FILTER.D3D11_FILTER_MIN_MAG_MIP_LINEAR;
		samplerDesc.AddressU = D3D11_TEXTURE_ADDRESS_MODE.D3D11_TEXTURE_ADDRESS_WRAP;
		samplerDesc.AddressV = D3D11_TEXTURE_ADDRESS_MODE.D3D11_TEXTURE_ADDRESS_WRAP;
		samplerDesc.AddressW = D3D11_TEXTURE_ADDRESS_MODE.D3D11_TEXTURE_ADDRESS_WRAP;
		samplerDesc.MipLODBias = 0.0f;
		samplerDesc.MaxAnisotropy = 1;
		samplerDesc.ComparisonFunc = D3D11_COMPARISON_FUNC.D3D11_COMPARISON_ALWAYS;
		samplerDesc.BorderColor[0] = 0.0f;
		samplerDesc.BorderColor[1] = 0.0f;
		samplerDesc.BorderColor[2] = 0.0f;
		samplerDesc.BorderColor[3] = 0.0f;
		samplerDesc.MinLOD = 0.0f;
		samplerDesc.MaxLOD = D3D11_FLOAT32_MAX;

		if (FAILED(device_.CreateSamplerState(&samplerDesc, &samplerState_)))
		{
			cleanup();
		    throw new Error("Failed create sampler state");
		}


		//--------------------------------------------------------------
		// Output Merger ステージで使用するデータを準備

		// バックバッファを2Dテクスチャとして取得
		ID3D11Texture2D backBuffer;
		swapChain_.GetBuffer(0, &IID_ID3D11Texture2D, cast(void**)&backBuffer);

		// バックバッファをレンダリングターゲットとするインターフェースを作成
		if (FAILED(device_.CreateRenderTargetView(backBuffer, null, &renderTargetView_)))
		{
			backBuffer.Release();
			cleanup();
			throw new Error("Failed create render target view");
		}

		backBuffer.Release();


		// 深度ステンシルバッファの設定
		D3D11_DEPTH_STENCIL_DESC depthStencilDesc;
		depthStencilDesc.DepthEnable = TRUE;
		depthStencilDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK.D3D11_DEPTH_WRITE_MASK_ALL;
		depthStencilDesc.DepthFunc = D3D11_COMPARISON_FUNC.D3D11_COMPARISON_LESS_EQUAL;
		depthStencilDesc.StencilEnable = FALSE;
		if (FAILED(device_.CreateDepthStencilState(&depthStencilDesc, &depthStencilState_)))
		{
			cleanup();
			throw new Error("Failed create depth stencil state");
		}

		// アルファブレンドの設定
		D3D11_BLEND_DESC blendStateDesc;
		blendStateDesc.AlphaToCoverageEnable = FALSE;
		blendStateDesc.IndependentBlendEnable = FALSE;
		blendStateDesc.RenderTarget[0].BlendEnable = TRUE;
		blendStateDesc.RenderTarget[0].SrcBlend = D3D11_BLEND.D3D11_BLEND_SRC_ALPHA;
		blendStateDesc.RenderTarget[0].DestBlend = D3D11_BLEND.D3D11_BLEND_INV_SRC_ALPHA;
		blendStateDesc.RenderTarget[0].BlendOp = D3D11_BLEND_OP.D3D11_BLEND_OP_ADD;
		blendStateDesc.RenderTarget[0].SrcBlendAlpha = D3D11_BLEND.D3D11_BLEND_SRC_ALPHA;
		blendStateDesc.RenderTarget[0].DestBlendAlpha = D3D11_BLEND.D3D11_BLEND_INV_SRC_ALPHA;
		blendStateDesc.RenderTarget[0].BlendOpAlpha = D3D11_BLEND_OP.D3D11_BLEND_OP_ADD;
		blendStateDesc.RenderTarget[0].RenderTargetWriteMask = D3D11_COLOR_WRITE_ENABLE.D3D11_COLOR_WRITE_ENABLE_ALL;
		if (FAILED(device_.CreateBlendState(&blendStateDesc, &blendState_)))
		{
			cleanup();
			throw new Error("Failed create blend state");
		}


		// 深度データ生成用処理
		try
		{
			depthRenderTechnique_ = new DepthRenderTechnique(device_, D3DCompile, 2048, 2048);
		}
		catch (Error e)
		{
			cleanup();
			throw new Error("Failed create depth render technique");
		}


		// モデルデータをロード
		pmdFile_ = new PmdFile(filePath, device_);

		// モデルの中心点を原点に合わせるようワールド座標変換行列をセット
		Vector3 min = Vector3(0, 0, 0);
		Vector3 max = Vector3(0, 0, 0);
		foreach (vertex; pmdFile_.vertices)
		{
			if (max.x < vertex.x)
				max.x = vertex.x;
			if (max.y < vertex.y)
				max.y = vertex.y;
			if (max.z < vertex.z)
				max.z = vertex.z;
			if (min.x > vertex.x)
				min.x = vertex.x;
			if (min.y > vertex.y)
				min.y = vertex.y;
			if (min.z > vertex.z)
				min.z = vertex.z;
		}
		constants_.worldMatrix = createTranslationMatrix(-(max.x + min.x) / 2, -(max.y + min.y) / 2, -(max.z + min.z) / 2);

		// カメラ座標を初期化
		camera_.x = 0;
		camera_.y = 0;
		camera_.z = -30;

		// 入力デバイス初期化
		mouse_ = Input.instance.mouse;
	}

	/++
	 + デストラクタ
	 +
	 + アプリケーション終了時に呼び出される
	 +/
	~this()
	{
		cleanup();
	}

	/++
	 + 毎フレームの処理
	 +
	 + 一定時間毎に呼び出される
	 +/
	void mainLoop()
	{
		// マウス入力状態を更新
		Input.instance.refresh();

		// 左クリック時にマウス移動でモデルを回転
		if (mouse_.isPress(Mouse.Button.LEFT))
		{
			if (mouse_.moveX != 0)
			{
				Matrix rotation = createYRotationMatrix(-0.01 * cast(float)mouse_.moveX);
				constants_.worldMatrix = constants_.worldMatrix * rotation;
			}

			if (mouse_.moveY != 0)
			{
				Matrix rotation = createXRotationMatrix(0.01 * cast(float)mouse_.moveY);
				constants_.worldMatrix = constants_.worldMatrix * rotation;
			}
		}

		// マウスホイールでモデルまでの距離を変更
		camera_.z += mouse_.moveZ * 0.01;

		// 視線の対象
		Vector3 target = Vector3(0, 0, 0);

		// カメラ上方向ベクトル
		Vector3 cameraUp = Vector3(0, 1, 0);

		// ライト原点
		Vector3 light = Vector3(0, 30, -30);

		// ライトをカメラとしたときのビュー変換行列を生成
		Matrix lightViewMatrix = createViewMatrix(light, cameraUp, target);

		// 深度データ生成は正射影変換を使用する(範囲は仮で直指定)
		Matrix orthogonalProjectionMatrix = createOrthogonalProjectionMatrix(0, 100, 30, 30);

		// 深度データ生成用にワールドビュー射影変換行列を生成
		constants_.worldLightViewProjectionMatrix = constants_.worldMatrix * lightViewMatrix * orthogonalProjectionMatrix;


		// カメラからのビュー行列を生成
		Matrix cameraViewMatrix = createViewMatrix(camera_, cameraUp, target);

		// 通常レンダリング用の射影変換
		Matrix projectionMatrix = createProjectionMatrix(PI / 180 * 60, 0.5, 250, Application.instance.window.width, Application.instance.window.height);

		// 通常レンダリング用のワールドビュー射影変換行列
		constants_.worldCameraViewProjectionMatrix = constants_.worldMatrix * cameraViewMatrix * projectionMatrix;


		//--------------------------------------------------------------
		// 深度データをレンダリング
		ID3D11ShaderResourceView depthTextureView = depthRenderTechnique_.render(immediateContext_, pmdFile_, constants_.worldLightViewProjectionMatrix);


		//--------------------------------------------------------------
		// パイプラインの各ステージの設定

		// Input Assember ステージに入力するデータ形式を設定
		immediateContext_.IASetInputLayout(inputLayout_);

		// Input Assember ステージに入力されたデータの使用方法を設定
		immediateContext_.IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY.D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST);

		// Vertex Shader ステージに Vertex Shader を設定
		immediateContext_.VSSetShader(vertexShader_, null, 0);

		// 定数バッファの内容を更新
		immediateContext_.UpdateSubresource(vertexShaderConstantsBuffer_, 0, null, &constants_, 0, 0);
		
		// Vertex Shaderに定数バッファをセット
		immediateContext_.VSSetConstantBuffers(0, 1, &vertexShaderConstantsBuffer_);

		// Rasterizer ステージにビューポートを設定
		immediateContext_.RSSetViewports(1, &viewPort_);

		// Rasterizer ステージの設定
		immediateContext_.RSSetState(rasterizerState_);

		// Output Merger ステージで最終的なレンダリングターゲットをバックバッファに設定
		immediateContext_.OMSetRenderTargets(1, &renderTargetView_, depthStencilView_);

		// Output Merger ステージで深度ステンシルを設定
		immediateContext_.OMSetDepthStencilState(depthStencilState_, 0);

		// Output Merger ステージでアルファブレンドを設定
		immediateContext_.OMSetBlendState(blendState_, null, 0xffffffff);


		//--------------------------------------------------------------
		// 描画処理

		// 深度バッファをクリア
		immediateContext_.ClearDepthStencilView(depthStencilView_, D3D11_CLEAR_FLAG.D3D11_CLEAR_DEPTH, 1.0f, 0);

		// レンダリングターゲットを一色で塗りつぶし
		float[4] clearColor = [0, 0, 1, 0];
		immediateContext_.ClearRenderTargetView(renderTargetView_, clearColor.ptr);

		// Input Assembler ステージに頂点をセット
		uint stride = PmdFile.Vertex.sizeof;
		uint offset = 0;

		immediateContext_.IASetVertexBuffers(0, 1, &pmdFile_.vertexBuffer, &stride, &offset);

		for (uint i; i < pmdFile_.parts.length; ++i)
		{
			// Input Assembler ステージにインデックスをセット
			immediateContext_.IASetIndexBuffer(pmdFile_.parts[i].indexBuffer, DXGI_FORMAT.DXGI_FORMAT_R16_UINT, 0);

			// テクスチャ有無でシェーダを分ける
			if (pmdFile_.parts[i].textureView is null)
			{
				// Pixel Shaderに Pixel Shader を設定
				immediateContext_.PSSetShader(polygonPixelShader_, null, 0);
			}
			else
			{
				// Pixel Shaderに Pixel Shader を設定
				immediateContext_.PSSetShader(texturePixelShader_, null, 0);

				// Pixel Shaderにテクスチャをセット
				immediateContext_.PSSetShaderResources(0, 1, &pmdFile_.parts[i].textureView);
			}

			// 深度データをt1にセット
			immediateContext_.PSSetShaderResources(1, 1, &depthTextureView);

			// Pixel Shaderにサンプラをセット
			immediateContext_.PSSetSamplers(0, 1, &samplerState_);

			// Pixel Shaderの定数バッファの内容を更新
			immediateContext_.UpdateSubresource(pixelShaderConstantsBuffer_, 0, null, &pmdFile_.parts[i].material, 0, 0);

			// Pixel Shaderに定数バッファをセット
			immediateContext_.PSSetConstantBuffers(0, 1, &pixelShaderConstantsBuffer_);

			// パイプラインを実行して描画する
			immediateContext_.DrawIndexed(pmdFile_.parts[i].vertexCount, 0, 0);
		}

		// 次にテクスチャにレンダリングするために、インプットからは外して置かなければならない
		ID3D11ShaderResourceView nullView = null;
		immediateContext_.PSSetShaderResources(1, 1, &nullView);

		// バックバッファを表示
		swapChain_.Present(0, 0);
	}

private:

	// d3d11.dllのハンドル
	HMODULE d3d11Dll_;

	// d3dcompiler_xx.dllのハンドル
	HMODULE d3dcompiler_;

	// Swap chain
	IDXGISwapChain swapChain_;

	// Device
	ID3D11Device device_;

	// Device context
	ID3D11DeviceContext immediateContext_;

	// レンダリングターゲット
	ID3D11RenderTargetView renderTargetView_;

	// 深度バッファ
	ID3D11DepthStencilView depthStencilView_;

	// Input layout
	ID3D11InputLayout inputLayout_;

	// Vertex shader
	ID3D11VertexShader vertexShader_;

	// 頂点シェーダ定数バッファ
	ID3D11Buffer vertexShaderConstantsBuffer_;

	// View port
	D3D11_VIEWPORT viewPort_;

	// Rasterizer設定
	ID3D11RasterizerState rasterizerState_;

	// テクスチャを使用しないPixel shader
	ID3D11PixelShader polygonPixelShader_;

	// テクスチャを使用するPixel shader
	ID3D11PixelShader texturePixelShader_;

	// ピクセルシェーダ定数バッファ
	ID3D11Buffer pixelShaderConstantsBuffer_;

	// サンプラ
	ID3D11SamplerState samplerState_;

	// 深度ステンシル設定
	ID3D11DepthStencilState depthStencilState_;

	// アルファブレンド設定
	ID3D11BlendState blendState_;

	// 深度バッファシャドウの深度テクスチャ生成用パイプライン
	DepthRenderTechnique depthRenderTechnique_;

	// モデルデータ
	PmdFile pmdFile_;

	// 定数バッファに送るデータ
	Constants constants_;

	// モデルとカメラの距離
	Vector3 camera_;

	// マウス
	const(Mouse) mouse_;


	/+
	 + 解放処理
	 +/
	void cleanup()
	{
		if (pmdFile_ !is null)
		{
			delete pmdFile_;
			pmdFile_ = null;
		}

		if (depthRenderTechnique_ !is null)
		{
			delete depthRenderTechnique_;
			depthRenderTechnique_ = null;
		}

		if (blendState_ !is null)
		{
			blendState_.Release();
			blendState_ = null;
		}

		if (depthStencilState_ !is null)
		{
			depthStencilState_.Release();
			depthStencilState_ = null;
		}

		if (samplerState_ !is null)
		{
			samplerState_.Release();
			samplerState_ = null;
		}

		if (pixelShaderConstantsBuffer_ !is null)
		{
			pixelShaderConstantsBuffer_.Release();
			pixelShaderConstantsBuffer_ = null;
		}

		if (texturePixelShader_ !is null)
		{
			texturePixelShader_.Release();
			texturePixelShader_ = null;
		}

		if (polygonPixelShader_ !is null)
		{
			polygonPixelShader_.Release();
			polygonPixelShader_ = null;
		}

		if (rasterizerState_ !is null)
		{
			rasterizerState_.Release();
			rasterizerState_ = null;
		}

		if (vertexShaderConstantsBuffer_ !is null)
		{
			vertexShaderConstantsBuffer_.Release();
			vertexShaderConstantsBuffer_ = null;
		}

		if (vertexShader_ !is null)
		{
			vertexShader_.Release();
			vertexShader_ = null;
		}

		if (inputLayout_ !is null)
		{
			inputLayout_.Release();
			inputLayout_ = null;
		}

		if (depthStencilView_ !is null)
		{
			depthStencilView_.Release();
			depthStencilView_ = null;
		}

		if (renderTargetView_ !is null)
		{
			renderTargetView_.Release();
			renderTargetView_ = null;
		}

		if (immediateContext_ !is null)
		{
			immediateContext_.Release();
			immediateContext_ = null;
		}

		if (device_ !is null)
		{
			device_.Release();
			device_ = null;
		}

		if (swapChain_ !is null)
		{
			swapChain_.Release();
			swapChain_ = null;
		}

		if (d3dcompiler_ !is null)
		{
			FreeLibrary(d3dcompiler_);
			d3dcompiler_ = null;
		}

		if (d3d11Dll_ !is null)
		{
			FreeLibrary(d3d11Dll_);
			d3d11Dll_ = null;
		}
	}
}

